"""Set of log handlers for sending logs to Loki."""

# Copied from https://github.com/RomanR-dev/python-logging-loki

import logging
import warnings
from logging.handlers import QueueHandler, QueueListener
from queue import Queue
from typing import Dict, Optional, Type

from chime_logger import const
from chime_logger import emitters as emitter


class LokiQueueHandler(QueueHandler):
    """This handler automatically creates listener and `LokiHandler` to handle logs queue."""

    def __init__(self, queue: Queue, **kwargs):
        """Create new logger handler with the specified queue and kwargs for the `LokiHandler`."""
        super().__init__(queue)
        self.handler = LokiHandler(**kwargs)  # noqa: WPS110
        self.listener = QueueListener(self.queue, self.handler)
        self.listener.start()


class LokiHandler(logging.Handler):
    """Log handler that sends log records to Loki.

    `Loki API <https://github.com/grafana/loki/blob/master/docs/api.md>`_
    """

    emitters: Dict[str, Type[emitter.LokiEmitter]] = {
        "0": emitter.LokiEmitterV0,
        "1": emitter.LokiEmitterV1,
        "2": emitter.LokiEmitterV2,
    }

    def __init__(
        self,
        url: str,
        tags: Optional[dict] = None,
        auth: Optional[emitter.BasicAuth] = None,
        version: Optional[str] = None,
        headers: Optional[dict] = None,
    ):
        """Create new Loki logging handler.

        Arguments:
            url: Endpoint used to send log entries to Loki
                 (e.g. `https://my-loki-instance/loki/api/v1/push`).
            tags: Default tags added to every log record.
            auth: Optional tuple with username and password for basic HTTP authentication.
            version: Version of Loki emitter to use.
            headers: HTTP headers to send with request.
        """
        super().__init__()

        if version is None and const.emitter_ver == "0":
            msg = (
                "Loki /api/prom/push endpoint is in the depreciation process starting from version 0.4.0.",  # noqa: E501
                "Explicitly set the emitter version to '0' if you want to use the old endpoint.",
                "Or specify '1' if you have Loki version> = 0.4.0.",
                "When the old API is removed from Loki, the handler will use the new version by default.",  # noqa: E501
            )
            warnings.warn(" ".join(msg), DeprecationWarning)

        version = version or const.emitter_ver
        if version not in self.emitters:
            raise ValueError("Unknown emitter version: {}".format(version))
        self.emitter = self.emitters[version](url, tags, auth, headers)

    def handleError(self, record):  # noqa: N802
        """Close emitter and let default handler take actions on error."""
        self.emitter.close()
        super().handleError(record)

    def emit(self, record: logging.LogRecord):
        """Send log record to Loki."""
        # noinspection PyBroadException
        try:
            self.emitter(record, self.format(record))
        except Exception:
            self.handleError(record)



"""Loki emitters for sending logs to Loki."""

# Copied from https://github.com/RomanR-dev/python-logging-loki

import abc
import copy
import functools
import logging
import time
from logging.config import ConvertingDict
from typing import Any, Dict, List, Optional, Tuple

import requests
import rfc3339

from chime_logger import const

BasicAuth = Optional[Tuple[str, str]]


class LokiEmitter(abc.ABC):
    """Base Loki emitter class."""

    success_response_code = const.success_response_code
    level_tag = const.level_tag
    logger_tag = const.logger_tag
    label_allowed_chars = const.label_allowed_chars
    label_replace_with = const.label_replace_with
    session_class = requests.Session

    def __init__(
        self,
        url: str,
        tags: Optional[Dict[str, Any]] = None,
        auth: BasicAuth = None,
        headers: Optional[Dict[str, Any]] = None,
    ):
        """Create new Loki emitter.

        Arguments:
            url: Endpoint used to send log entries to Loki
                 (e.g. `https://my-loki-instance/loki/api/v1/push`).
            tags: Default tags added to every log record.
            auth: Optional tuple with username and password for basic HTTP authentication.
            headers: Optional dict with HTTP headers to send.
        """
        #: Tags that will be added to all records handled by this handler.
        self.tags = tags or {}
        #: Loki JSON push endpoint (e.g `http://127.0.0.1/loki/api/v1/push`)
        self.url = url
        #: Optional tuple with username and password for basic authentication.
        self.auth = auth
        #: Optional headers for post request
        self.headers = headers or {}

        self._session: Optional[requests.Session] = None

    def __call__(self, record: logging.LogRecord, line: str):
        """Send log record to Loki."""
        payload = self.build_payload(record, line)
        resp = self.session.post(self.url, json=payload, headers=self.headers)
        if resp.status_code != self.success_response_code:
            raise ValueError(
                "Unexpected Loki API response status code: {}".format(resp.status_code)
            )

    @abc.abstractmethod
    def build_payload(self, record: logging.LogRecord, line) -> dict:
        """Build JSON payload with a log entry."""
        raise NotImplementedError  # pragma: no cover

    @property
    def session(self) -> requests.Session:
        """Create HTTP session."""
        if self._session is None:
            self._session = self.session_class()
            self._session.auth = self.auth or None
        return self._session

    def close(self):
        """Close HTTP session."""
        if self._session is not None:
            self._session.close()
            self._session = None

    @functools.lru_cache(const.format_label_lru_size)
    def format_label(self, label: str) -> str:
        """Build label to match prometheus format.

        `Label format <https://prometheus.io/docs/concepts/data_model/#metric-names-and-labels>`_
        """
        for char_from, char_to in self.label_replace_with:
            label = label.replace(char_from, char_to)
        return "".join(char for char in label if char in self.label_allowed_chars)

    def build_tags(self, record: logging.LogRecord) -> Dict[str, Any]:
        """Return tags that must be send to Loki with a log record."""
        tags = dict(self.tags) if isinstance(self.tags, ConvertingDict) else self.tags
        tags = copy.deepcopy(tags)
        tags[self.level_tag] = record.levelname.lower()
        tags[self.logger_tag] = record.name

        extra_tags = getattr(record, "tags", {})
        if not isinstance(extra_tags, dict):
            return tags

        for tag_name, tag_value in extra_tags.items():
            cleared_name = self.format_label(tag_name)
            if cleared_name:
                tags[cleared_name] = tag_value

        return tags


class LokiEmitterV0(LokiEmitter):
    """Emitter for Loki < 0.4.0."""

    def build_payload(self, record: logging.LogRecord, line) -> dict:
        """Build JSON payload with a log entry."""
        labels = self.build_labels(record)
        ts = rfc3339.format_microsecond(record.created)
        stream = {
            "labels": labels,
            "entries": [{"ts": ts, "line": line}],
        }
        return {"streams": [stream]}

    def build_labels(self, record: logging.LogRecord) -> str:
        """Return Loki labels string."""
        labels: List[str] = []
        for label_name, label_value in self.build_tags(record).items():
            cleared_name = self.format_label(str(label_name))
            cleared_value = str(label_value).replace('"', r"\"")
            labels.append('{}="{}"'.format(cleared_name, cleared_value))
        return "{{{0}}}".format(",".join(labels))


class LokiEmitterV1(LokiEmitter):
    """Emitter for Loki >= 0.4.0."""

    def build_payload(self, record: logging.LogRecord, line) -> dict:
        """Build JSON payload with a log entry."""
        labels = self.build_tags(record)
        ns = 1e9
        ts = str(int(time.time() * ns))
        stream = {
            "stream": labels,
            "values": [[ts, line]],
        }
        return {"streams": [stream]}


class LokiEmitterV2(LokiEmitterV1):
    """Emitter for Loki >= 0.4.0.

    Enables passing additional headers to requests
    """

    def __init__(
        self,
        url: str,
        tags: Optional[dict] = None,
        auth: BasicAuth = None,
        headers: Optional[dict] = None,
    ):
        """Create new V2 Loki emitter.

        Arguments:
            url: Endpoint used to send log entries to Loki
                 (e.g. `https://my-loki-instance/loki/api/v1/push`).
            tags: Default tags added to every log record.
            auth: Optional tuple with username and password for basic HTTP authentication.
            headers: Optional dict with HTTP headers to send.
        """
        super().__init__(url, tags, auth, headers)

    def __call__(self, record: logging.LogRecord, line: str):
        """Send log record to Loki."""
        payload = self.build_payload(record, line)
        resp = self.session.post(self.url, json=payload, headers=self.headers)
        if resp.status_code != self.success_response_code:
            raise ValueError(
                "Unexpected Loki API response status code: {}".format(resp.status_code)
            )



"""Configuration for the CHIME logging system."""

import os
import pathlib

DEFAULT_LOKI_URL = "https://frb.chimenet.ca/loki/loki/api/v1/push"
DEFAULT_LOKI_TENANT = "CHIME"
DEFAULT_FILE_LOG_PATH = "logs/my_app.logs"

LOKI_AUTH = (
    (
        os.getenv("CHIME_LOGGER_LOKI_USER"),
        os.getenv("CHIME_LOGGER_LOKI_PASSWORD"),
    )
    if os.getenv("CHIME_LOGGER_LOKI_USER") and os.getenv("CHIME_LOGGER_LOKI_PASSWORD")
    else None
)


def check_file_log_path_valid(path: str) -> str:
    """Check if the file log path is valid and return the absolute path.

    Args:
        path (str): The file log path to check.

    Returns:
        str: The absolute path of the file log.
    """
    abs_path = pathlib.Path(path).expanduser().resolve()
    if not abs_path.parent.exists():
        abs_path.parent.mkdir(parents=True, exist_ok=True)
    return str(abs_path)


# TODO: Allow pushing to multiple Loki instances
LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simple": {"format": "[ pipeline=%(pipeline)s event=%(event)s ] %(message)s"},
        "detailed": {
            "format": "%(asctime)s [ pipeline=%(pipeline)s event=%(event)s ] %(message)s",
            "datefmt": "%Y-%m-%dT%H:%M:%S%z",
        },
    },
    "filters": {
        "add_pipeline_filter": {"()": "chime_logger.filters.PipelineFilter"},
        "add_event_filter": {"()": "chime_logger.filters.EventFilter"},
    },
    "handlers": {
        "loki": {
            "()": "chime_logger.handlers.LokiHandler",
            "level": "INFO",
            "formatter": "simple",
            "filters": ["add_pipeline_filter", "add_event_filter"],
            "url": os.getenv("CHIME_LOGGER_LOKI_URL", DEFAULT_LOKI_URL),
            "auth": LOKI_AUTH,
            "headers": {
                "X-Scope-OrgID": os.getenv(
                    "CHIME_LOGGER_LOKI_TENANT", DEFAULT_LOKI_TENANT
                )
            },
            "version": "2",
        },
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "INFO",
            "formatter": "detailed",
            # BUG: This can fail if the directory does not exist
            "filename": check_file_log_path_valid(
                os.getenv("CHIME_LOGGER_FILE_LOG_PATH", DEFAULT_FILE_LOG_PATH)
            ),
            "maxBytes": 10_000_000,
            "backupCount": 3,
        },
        "queue_handler": {
            "class": "logging.handlers.QueueHandler",
            "handlers": ["loki", "file"],
            "filters": ["add_pipeline_filter", "add_event_filter"],
            "respect_handler_level": True,
        },
    },
    "loggers": {
        "CHIME": {"level": "INFO", "handlers": ["queue_handler"], "propagate": False}
    },
}
